/*****************************************************************************
#                                                                            #
#    uStreamer - Lightweight and fast MJPG-HTTP streamer.                    #
#                                                                            #
#    Copyright (C) 2018  Maxim Devaev <mdevaev@gmail.com>                    #
#                                                                            #
#    This program is free software: you can redistribute it and/or modify    #
#    it under the terms of the GNU General Public License as published by    #
#    the Free Software Foundation, either version 3 of the License, or       #
#    (at your option) any later version.                                     #
#                                                                            #
#    This program is distributed in the hope that it will be useful,         #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of          #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
#    GNU General Public License for more details.                            #
#                                                                            #
#    You should have received a copy of the GNU General Public License       #
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.  #
#                                                                            #
*****************************************************************************/


#include "stream.h"

#include <stdlib.h>
#include <stdbool.h>
#include <stdatomic.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#include <pthread.h>

#include "tools.h"
#include "threading.h"
#include "logging.h"
#include "picture.h"
#include "device.h"
#include "encoder.h"
#ifdef WITH_GPIO
#	include "gpio.h"
#endif


//#define WITH_BLANKBUF
//#define WITH_FB
//#define HTTP_GETRAW

#ifdef WITH_FB
  #include <linux/fb.h>
  #include <fcntl.h>
  #include <sys/ioctl.h>
  #include <sys/mman.h>
  #define COPYFB(_next) copyfb(_next)
  #define COPYFB2(_next)
#else
  #define COPYFB(_next)
#endif

#ifdef HTTP_GETRAW
  #include "http/server.h"
#endif

//#if ! defined xxx
//  #pragma message "test msg"
//#endif

#define EMPTY

struct _worker_t {
	pthread_t			tid;
	unsigned			number;
	atomic_bool			*proc_stop;
	atomic_bool			*workers_stop;

	long double			last_comp_time;

	pthread_mutex_t		has_job_mutex;
	unsigned			buf_index;
	atomic_bool			has_job;
	bool				job_timely;
	bool				job_failed;
	long double			job_start_ts;
	pthread_cond_t		has_job_cond;

	pthread_mutex_t		*free_workers_mutex;
	unsigned			*free_workers;
	pthread_cond_t		*free_workers_cond;

	struct _worker_t	*order_prev;
	struct _worker_t	*order_next;

	struct device_t		*dev;
	struct encoder_t	*encoder;
};

struct _workers_pool_t {
	unsigned			n_workers;
	struct _worker_t	*workers;
	struct _worker_t	*oldest_worker;
	struct _worker_t	*latest_worker;

	long double			approx_comp_time;

	pthread_mutex_t		free_workers_mutex;
	unsigned			free_workers;
	pthread_cond_t		free_workers_cond;

	atomic_bool			workers_stop;

	long double			desired_frames_interval;
};


static struct _workers_pool_t *_stream_init_loop(struct stream_t *stream);
static struct _workers_pool_t *_stream_init(struct stream_t *stream);
static void _stream_expose_picture(struct stream_t *stream, unsigned buf_index, unsigned captured_fps);

static struct _workers_pool_t *_workers_pool_init(struct stream_t *stream);
static void _workers_pool_destroy(struct _workers_pool_t *pool);

static void *_worker_thread(void *v_worker);

static struct _worker_t *_workers_pool_wait(struct _workers_pool_t *pool);
static void _workers_pool_assign(struct _workers_pool_t *pool, struct _worker_t *ready_worker, unsigned buf_index);
static long double _workers_pool_get_fluency_delay(struct _workers_pool_t *pool, struct _worker_t *ready_worker);


struct stream_t *stream_init(struct device_t *dev, struct encoder_t *encoder) {
	struct process_t *proc;
	struct stream_t *stream;

	A_CALLOC(proc, 1);
	atomic_init(&proc->stop, false);
	atomic_init(&proc->slowdown, false);

	A_CALLOC(stream, 1);
	stream->picture = picture_init();
	stream->dev = dev;
	stream->encoder = encoder;
	atomic_init(&stream->updated, false);
	A_MUTEX_INIT(&stream->mutex);
	stream->proc = proc;

//        stream->reconnect = false;
//        stream->raw_request = false;
//        stream->raw_done = false;
	return stream;
}

void stream_destroy(struct stream_t *stream) {
	A_MUTEX_DESTROY(&stream->mutex);
	picture_destroy(stream->picture);
	free(stream->proc);
	free(stream);
}






int draw_Y=0;
int draw_U=0x80;
int draw_V=0x80;
int draw_fontdbl=2;
int draw_xres=0;
int draw_yres=0;
//int draw_maxbu
int real_fps=0;
int text_x=0;
int text_y=0;
int draw_bpp=2;  // bytes per pixel, 2 for YUV
char*device_path;  // local copy of dev->path, dirty hack

int raspicam_rightmost_correction=0;


#ifdef WITH_FB
char*fb_dev=NULL;
unsigned char*fbptr=NULL;
int fbfd=0;
int fbx=320; // TODO: autodetect
int fbxx=640;
int fby=240;
int fb_on=0;
int fb_bpp=0;
//char fb_dev[128]="/dev/fb1";
struct fb_fix_screeninfo fb_finfo;
struct fb_var_screeninfo fb_vinfo;
long double             lastfb_ts=0;
long double fb_mininterval=0.08; // minimum interval, in seconds, for display refresh; 0.08=12.5 fps; keep low for slow SPI displays

void fb_open(char*fn){
  if(fb_dev==NULL){LOG_INFO("no framebuffer -F option specified, local display echo disabled");return;}
  fbfd=open(fn,O_RDWR);
  if(fbfd<0){
    LOG_ERROR("ERROR: can not open display '%s', local display echo disabled",fn);
    return;}
  ioctl(fbfd, FBIOGET_VSCREENINFO, &fb_vinfo);
  ioctl(fbfd, FBIOGET_FSCREENINFO, &fb_finfo);
  fbx=fb_vinfo.xres;
  fby=fb_vinfo.yres;
  fbxx=fb_finfo.line_length;
  //fb_bpp=fb_vinfo.bits_per_pixel;
  fb_bpp=fb_finfo.line_length/fbx; // 2 or 3 or 4
  fb_on=1;
  LOG_INFO("Opening local display '%s': res %d x %d, %d bpp; %d bytes/line",fn,fbx,fby, fb_vinfo.bits_per_pixel, fbxx);

  // set mmap
  fbptr = mmap(0, fb_vinfo.yres * fb_finfo.line_length, PROT_READ | PROT_WRITE, MAP_SHARED, fbfd, 0);
}



void fb_close(){
  if(fb_on)close(fbfd);
  fb_on=0;
}

#endif
#ifdef WITH_FB

void setfbpix_YUV422_RGB565(unsigned char*fbufpos,unsigned char*cbufpos){
  int r,g,b;
// r = y + 1.14 v               // 368 * v
// g = y - .395 u - 0.581 v     // 101 * u - 149 * v
// b = y + 2.033 u              // 520 * u
  int uu,vv,y;
  y=cbufpos[0]<<8;
  uu=cbufpos[1]-128;
  vv=cbufpos[3]-128;
  r=y + 368*vv;
  g=y - 101*uu - 149*vv;
  b=y + 520*uu;
  if(r<0)r=0; else if(r>65535)r=65535;
  if(g<0)g=0; else if(g>65535)g=65535;
  if(b<0)b=0; else if(b>65535)b=65535;
  r>>=(8+3);r&=0x1f;
  g>>=(8+2);g&=0x3f;
  b>>=(8+3);b&=0x1f;
  fbufpos[0]=b+(g<<5);
  fbufpos[1]=(r<<3)+(g>>3);
}

void copyfb(unsigned char* buf){
  int x,y;
  unsigned char*fbuf=fbptr;
  unsigned char*cbuf=buf;
  unsigned char*ycbuf=buf;
  long double ts;
  double dxf,dyf;
  if(!fb_on)return; // no fb enabled, fail silently
  //LOG_INFO("copyfb: %.3Lf",get_now_monotonic()-lastfb_ts);
  if(get_now_monotonic()-lastfb_ts < fb_mininterval)return; // refresh rate limiter
  //LOG_INFO("copyfb");

  dxf=draw_xres/fbx;
  dyf=draw_yres/fby;
  ts=get_now_monotonic();
  for(y=0;y<fby;y++){
    cbuf=buf+(int)floor(y*dyf*draw_xres*2);

    if(dxf==2) // optimized for two-to-one ratio
      for(x=0;x<fbx;x++){
        setfbpix_YUV422_RGB565(fbuf,cbuf);
        cbuf+=(draw_bpp<<1); // 2:1
        fbuf+=fb_bpp;
      }
    else // general scaling
      for(x=0;x<fbx;x++){
        cbuf=ycbuf+(((int)floor(x*dxf/2))<<2);
        setfbpix_YUV422_RGB565(fbuf,cbuf);
        fbuf+=fb_bpp;
      }


    cbuf+=2*640;
  }
  lastfb_ts=get_now_monotonic();
LOG_PERF("copyfb: %.3Lf",lastfb_ts-ts);
}

#endif



// from: https://github.com/basti79/LCD-fonts
//#include "LCD-fonts/10x16_horizontal_LSB_1.h"
//#include "LCD-fonts/10x16_horizontal_LSB_2.h"
//#include "LCD-fonts/12x16_horizontal_LSB_1.h"
//#include "LCD-fonts/12x16_horizontal_LSB_2.h"
//#include "LCD-fonts/12x20_horizontal_LSB_1.h"
//#include "LCD-fonts/12x20_horizontal_LSB_2.h"
//#include "LCD-fonts/16x26_horizontal_LSB_1.h"
//#include "LCD-fonts/16x26_horizontal_LSB_2.h"
//#include "LCD-fonts/22x36_horizontal_LSB_1.h"
//#include "LCD-fonts/22x36_horizontal_LSB_2.h"
//#include "LCD-fonts/24x40_horizontal_LSB_1.h"
//#include "LCD-fonts/24x40_horizontal_LSB_2.h"
//#include "LCD-fonts/32x53_horizontal_LSB_1.h"
//#include "LCD-fonts/32x53_horizontal_LSB_2.h"
//#include "LCD-fonts/4x6_horizontal_LSB_1.h"
//#include "LCD-fonts/4x6_horizontal_LSB_2.h"
//#include "LCD-fonts/5x12_horizontal_LSB_1.h"
//#include "LCD-fonts/5x12_horizontal_LSB_2.h"
//#include "LCD-fonts/5x8_horizontal_LSB_1.h"
//#include "LCD-fonts/5x8_horizontal_LSB_2.h"
//#include "LCD-fonts/6x10_horizontal_LSB_1.h"
//#include "LCD-fonts/6x10_horizontal_LSB_2.h"
//#include "LCD-fonts/6x8_horizontal_LSB_1.h"
//#include "LCD-fonts/6x8_horizontal_LSB_2.h"
//#include "LCD-fonts/7x12b_horizontal_LSB_1.h"
//#include "LCD-fonts/7x12b_horizontal_LSB_2.h"
//#include "LCD-fonts/7x12_horizontal_LSB_1.h"
//#include "LCD-fonts/7x12_horizontal_LSB_2.h"
//#include "LCD-fonts/8x12_horizontal_LSB_1.h"
//#include "LCD-fonts/8x12_horizontal_LSB_2.h"
//#include "LCD-fonts/8x14_horizontal_LSB_1.h"
//#include "LCD-fonts/8x14_horizontal_LSB_2.h"
//#include "LCD-fonts/8x8_horizontal_LSB_1.h"
//#include "LCD-fonts/8x8_horizontal_LSB_2.h"

#include "7x12_horizontal_LSB_2.h"

#define FONT_X 7
#define FONT_Y sizeof(font[0])
//#define FONT_XX (FONT_X+1)
#define FONT_XX (FONT_X)
#define FONT_YY (FONT_Y)
#define FONT_YLINES 2
#define FONT_YADJ 2

#define TEXT_PADX 3
#define TEXT_PADY 2

#define CLIP(X) ( (X) > 255 ? 255 : (X) < 0 ? 0 : X)
void draw_setcolor(int b,int g,int r){
  draw_Y=CLIP( ( 4096 + (r*76)  + (g*150) + (b*29)  ) >> 8);
  draw_U=( 32768 + (r*112) - (g*94) - (b*18)  ) >> 8;
  draw_V=( 32768 - (r*38)  - (g*74) + (b*112) ) >> 8;
}

void buffer_putpixelRGB(unsigned char *buf, int x,int y,int b,int g,int r){
  int pos;
  int cy,cu,cv;
//  x=x&0xFFFE; // zero last bit of coord
  x<<=1; // double x coord, we address in pixel pairs
  pos=(draw_xres*y+x)<<1; // linear buffer position
// https://stackoverflow.com/questions/1737726/how-to-perform-rgb-yuv-conversion-in-c-c/1737741#1737741
//
//  Y  =      (0.257 * R) + (0.504 * G) + (0.098 * B) + 16       = ( R<<6 + G<<7 + B*25 + 4096 ) >> 8
//  Cr = V =  (0.439 * R) - (0.368 * G) - (0.071 * B) + 128      = ( 32768 + R*112 - G*94 - B*18 ) >> 8
//  Cb = U = -(0.148 * R) - (0.291 * G) + (0.439 * B) + 128      = ( 32768 - R*38 - G*74 + B*112 ) >> 8

//  cy=(  4096 + (r<<6)  + (g<<7) + (b*25)  ) >> 8;
  cy=CLIP( ( 4096 + (r*76)  + (g*150) + (b*29)  ) >> 8);
  cu=( 32768 + (r*112) - (g*94) - (b*18)  ) >> 8;
  cv=( 32768 - (r*38)  - (g*74) + (b*112) ) >> 8;

  buf[pos  ]=cy; buf[pos+1]=cu; // Y:U
  buf[pos+2]=cy; buf[pos+3]=cv; // Y:V
}


void buffer_putpixel(unsigned char *buf, int x,int y){
  int pos;
  if(y>=draw_yres)return;
  x<<=1; // double x coord, we address in pixel pairs
  pos=(draw_xres*y+x)<<1; // linear buffer position
  buf[pos  ]=draw_Y; buf[pos+1]=draw_U; // Y:U
  buf[pos+2]=draw_Y; buf[pos+3]=draw_V; // Y:V
}


void buffer_putchar(unsigned char *buf, int c, int x, int y){
  if(c==' ')return; // space
  for(unsigned int yy=0;yy<FONT_Y;yy++){
    int rr=font[c][yy];
      if(rr)
      for(unsigned int xx=0;xx<FONT_X;xx++){
        if(rr&0x01){buffer_putpixel(buf,x+xx,y);
            if(draw_fontdbl>1)buffer_putpixel(buf,x+xx,y+1);}
        rr>>=1;
      }
      y++;
      if(draw_fontdbl)y++;
  }
}


/* void buffer_putcharw(unsigned char *buf, int c, int x, int y){
  for(unsigned int yy=0;yy<FONT_Y;yy++){
    for(int dbl=0;dbl<FONT_YLINES;dbl++){
      int rr=font[c][yy*2];
      for(unsigned int xx=0;xx<FONT_X;xx++){
        if(rr&0x01)buffer_putpixel(buf,x+xx,y);
//        if(rr&0x01)buffer_putpixel(buf,x+xx,y+1);
        rr>>=1;
      }
      rr=font[c][yy*2+1];
      for(unsigned int xx=0;xx<FONT_X;xx++){
        if(rr&0x01)buffer_putpixel(buf,x+xx,y);
//        if(rr&0x01)buffer_putpixel(buf,x+xx,y+1);
        rr>>=1;
      }
      y++;
    }
  }
} */


void buffer_putstr_raw(unsigned char *buf, int x,int y, char*s){
//  x>>=1;
  while(s[0]){
    buffer_putchar(buf,s[0],x,y);
//    buffer_putcharw(buf,s[0],x,y);
    s++;
    x+=FONT_XX;
  }
}

void buffer_putstr(unsigned char *buf, int x,int y, char*s){
  if(x<0)x=(draw_xres/2-TEXT_PADX/2+FONT_XX*x);
    else x=TEXT_PADX+FONT_XX*x;
  if(y<0)y=(draw_yres-TEXT_PADY+FONT_YADJ+FONT_YY*y);
    else y=TEXT_PADY+FONT_YADJ+FONT_YY*y;
  buffer_putstr_raw(buf,x,y,s);
//  while(s[0]){
//    buffer_putchar(buf,s[0],x,y);
//    buffer_putcharw(buf,s[0],x,y);
//    s++;
//    x+=FONT_XX;
//  }
}

void buffer_putstrr(unsigned char *buf, int x,int y, char*s){
  buffer_putstr(buf,-x-strlen(s),y,s);
}

void draw_darkblock(unsigned char*buf,int x1,int y1,int x2,int y2){
  for(int y=y1;y<y2;y++)for(int x=x1;x<x2;x++)buf[(draw_xres*y+x)*2]>>=2;
}


void buffer_puttime(unsigned char *buf, int x,int y){
  char s[256];
//  struct tm *tmi;
  struct timeval tv;
  time_t rawtime;
//  time( &rawtime );
//  tmi = localtime( &rawtime );
  gettimeofday(&tv,NULL);
  rawtime=tv.tv_sec;
  strftime(s,255,"%H:%M:%S",localtime(&rawtime));
  buffer_putstr(buf,x,y,s);
//  sprintf(s,".%06ld",tv.tv_usec);
  sprintf(s,".%03ld",(tv.tv_usec+500)/1000);
  buffer_putstr(buf,x+8,y,s);
  //fprintf(stderr,"Current local time and date: %s", asctime(tmi));
}

void draw_statusline(unsigned char *buf);





#ifdef WITH_BLANKBUF
unsigned char*blankbuf=NULL;
int blankbuf_size=0;
int blankbuf_x=0;
int blankbuf_y=0;
struct device_t *blankbuf_dev=NULL;
int blankbuf_ready=0;
int blankbuf_updated=0;
char*dev_errormsg=NULL;


void blankbuf_alloc(int x,int y){
  blankbuf_size=x*y*4; // *2 should be enough but let's be sure, memory is cheaper than debugging time
  A_CALLOC(blankbuf,blankbuf_size);
  blankbuf_x=x;
  blankbuf_y=y;
}

void blankbuf_realloc(int x,int y){
  int newsize=x*y*4;
  if(newsize>blankbuf_size){A_REALLOC(blankbuf,newsize);blankbuf_size=newsize;}
  blankbuf_x=x;
  blankbuf_y=y;
}

void blankbuf_clear(void){ // init YUYV
  int t;
  unsigned char*b=blankbuf;
  for(t=0;t<blankbuf_size/2;t++){
    b[0]=0;b++;
    b[0]=128;b++;
  }
}

void blankbuf_init(void){
  blankbuf_alloc(640,480);
  blankbuf_dev=device_init();
  blankbuf_dev->run->n_workers=1;
  blankbuf_dev->run->n_buffers=1;
  blankbuf_dev->run->width=640;
  blankbuf_dev->run->height=480;
  blankbuf_dev->run->format=blankbuf_dev->format;
  A_CALLOC(blankbuf_dev->run->hw_buffers, 1);
  blankbuf_dev->run->hw_buffers[0].data=blankbuf;
  blankbuf_dev->run->hw_buffers[0].allocated=blankbuf_x*blankbuf_y*2;
  _device_open_alloc_picbufs(blankbuf_dev);
}

void blankbuf_draw(void){
  blankbuf_clear();
  draw_xres=blankbuf_x;
  draw_yres=blankbuf_y;

//  buffer_putstr(blankbuf,0,1,"this is a test");
//  buffer_putstr(blankbuf,0,3,"this is a test");
//  buffer_putstr(blankbuf,0,5,"this is a test");
//  buffer_putstr(blankbuf,0,7,"this is a test");
  draw_setcolor(255,0,0);
  buffer_putstr(blankbuf, 10,7,"ERROR");
  buffer_putstr(blankbuf, 10,10,"Cannot open device");
  if(dev_errormsg)buffer_putstr(blankbuf, 10,16,dev_errormsg);
  draw_setcolor(192,128,0);
  buffer_putstr(blankbuf, 10,13,device_path);

  draw_statusline(blankbuf);

  blankbuf_dev->run->hw_buffers[0].used=blankbuf_x*blankbuf_y*2;
  blankbuf_updated=1;
}

void blankbuf_drawend(void){
  blankbuf_clear();
  draw_xres=blankbuf_x;
  draw_yres=blankbuf_y;
  draw_setcolor(255,0,0);
  buffer_putstr(blankbuf, 10,7,"STREAMING EXITED");
//  #ifdef WITH_FB
//  copyfb(blankbuf);
  COPYFB(blankbuf);
//  #endif
}

void blankbuf_showfb(void){
  blankbuf_draw();
//  #ifdef WITH_FB
  COPYFB(blankbuf);
//  #endif
}


#endif








int framecnt=0;




void draw_statusline(unsigned char *buf){
  char hostname[128];
//  char s[256];
  gethostname(hostname,127);
  draw_fontdbl=2;
  draw_darkblock(buf,0,draw_yres-FONT_YY*2-TEXT_PADY*2,draw_xres,draw_yres);

  draw_setcolor(255,255,0);
  buffer_putstr(buf,-strlen(device_path),-2,device_path);

  draw_setcolor(255,0,0);
  buffer_putstr(buf,-strlen(device_path)-1,-2,":");

  draw_setcolor(255,128,0);
  buffer_putstr(buf,-strlen(device_path)-strlen(hostname)-1,-2,hostname);

//  buffer_putstrr(buf, 0,-2,s);

  draw_setcolor(255,255,0); // time
  buffer_puttime(buf,0,-2); // 12 chars long
}


//void draw_into_buffer(struct stream_t *stream, int buf_index){
void draw_overlay(unsigned char*buf){

  text_x=(draw_xres-2*TEXT_PADX)/FONT_XX;
  text_y=(draw_yres-2*TEXT_PADY)/FONT_YY;

/*
  // test of whole-image mod (eg. for false colors, edge detect)
  if(false)
  for(int y=0;y<draw_yres;y++)for(int x=0;x<draw_xres/2;x++){
    unsigned char*pix;
    pix=buf+(draw_xres*y+x)*2;
    if(pix[0]<64)pix[0]<<=1;else
    pix[0]=(pix[0]>>1)+128;
  }

  if(false){
    for(int y=90;y<150;y++)for(int x=100;x<200;x++){buf[(draw_xres*y+x)*2]=0;buf[(draw_xres*y+x)*2+1]=0x80;} // zero Y and leave UV and get traces of transparency
    for(int y=90;y<150;y++)for(int x=300;x<400;x++){buf[(draw_xres*y+x)*2]>>=2;} // halve Y
    for(int y=90;y<150;y++)for(int x=100;x<150;x++)buffer_putpixelRGB(buf,x,y,x+y,x-y,0);
    buffer_putpixelRGB(buf,150/2,100,255,255,0);
  }

  if(false)
  for(int tt=0;tt<barw;tt++){
   int barw=10;
   int baroffs=200;
   for(int t=0;t<255;t++)buffer_putpixelRGB(buf,t,tt+baroffs       ,t,0,0);
   for(int t=0;t<255;t++)buffer_putpixelRGB(buf,t,tt+baroffs+  barw,0,t,0);
   for(int t=0;t<255;t++)buffer_putpixelRGB(buf,t,tt+baroffs+2*barw,0,0,t);
   for(int t=0;t<255;t++)buffer_putpixelRGB(buf,t,tt+baroffs+3*barw,t,t,0);
   for(int t=0;t<255;t++)buffer_putpixelRGB(buf,t,tt+baroffs+4*barw,t,0,t);
   for(int t=0;t<255;t++)buffer_putpixelRGB(buf,t,tt+baroffs+5*barw,0,t,t);
   for(int t=0;t<255;t++)buffer_putpixelRGB(buf,t,tt+baroffs+6*barw,t,t,t);
  }
//  draw_fontdbl=0;
//  buffer_putstr(buf,5,10,"this is a test");
//buffer_putstr(buf,0,0,"this is a test");
//  buffer_putstr(buf,5,30,"this is a test");
*/

  int r;
  char s[256];

  draw_setcolor(255,255,0);

  draw_statusline(buf);

  draw_setcolor(255,128,0);
  draw_fontdbl=2;

  snprintf(s,255,"%07d",framecnt);framecnt++;
  buffer_putstr(buf,0,-4, s);
  snprintf(s,255,"%4d",real_fps);
  buffer_putstrr(buf,0,-4, s);

#ifdef TEST_OVERLAY
  draw_fontdbl=1;
  draw_setcolor(0,255,0);
  r=rand();sprintf(s,"%010d",r);buffer_putstr(buf,-10,0,s);
  r=rand();sprintf(s,"%010d",r);buffer_putstr(buf,-10,2,s);
  r=rand();sprintf(s,"%010d",r);buffer_putstr(buf,-10,4,s);
  r=rand();sprintf(s,"%010d",r);buffer_putstr(buf,-10,6,s);
  r=rand();sprintf(s,"%010d",r);buffer_putstr(buf,-10,8,s);
  r=rand();sprintf(s,"%010d",r);buffer_putstr(buf,-10,10,s);

  draw_setcolor(0,255,255);
  buffer_putstr(buf,0,0,"this is a test");
  buffer_putstr(buf,0,2,"this is a test");
  buffer_putstr(buf,0,4,"this is a test");
  buffer_putstr(buf,0,6,"this is a test");
  buffer_putstr(buf,-1,12,"X");
  buffer_putstr(buf,-2,13,"X");
  buffer_putstr(buf,-3,14,"X");
  buffer_putstr(buf,-4,15,"X");
#endif
}


// get rid of raspi's annoying rightmost-pixel column
void draw_correctrightmost(unsigned char*buf,int xres,int yres){
  int xdiff=2*xres; // assume YUYV, 2 bytes per pixel
  // zero the rightmost
  //  int xpos=xres*2-2;for (int t=0;t<yres;t++){buf[xpos]=0;xpos+=xdiff;}
  // duplicate the rightmost from preceding
  //  int xpos=xres*2-4;for (int t=0;t<yres;t++){buf[xpos+2]=buf[xpos];xpos+=xdiff;}
  // preceding to last two pixels
  int xpos=xres*2-6;for (int t=0;t<yres;t++){buf[xpos+2]=buf[xpos];buf[xpos+4]=buf[xpos];xpos+=xdiff;}
}







void stream_loop(struct stream_t *stream) {
	struct _workers_pool_t *pool;

	#ifdef WITH_BLANKBUF
	 blankbuf_init();
	#endif
	#ifdef WITH_FB
	 fb_open(fb_dev);
	#endif

	LOG_INFO("Using V4L2 device: %s", stream->dev->path);
	LOG_INFO("Using desired FPS: %u", stream->dev->desired_fps);
	SEP_DEBUG('*');
	LOG_INFO("Starting main loop...");

	// main loop begins here; _stream_init_loop blocks when waiting for cam
	while ((pool = _stream_init_loop(stream)) != NULL) {
		long double grab_after = 0;
		unsigned fluency_passed = 0;
		unsigned captured_fps = 0;
		unsigned captured_fps_accum = 0;
		long long captured_fps_second = 0;
		bool persistent_timeout_reported = false;

                if(stream->dev->format != V4L2_PIX_FMT_YUYV){
  		    LOG_INFO("CAUTION: pixelformat not YUYV, can not overlay");
                }

		LOG_DEBUG("Pre-allocating memory for stream picture ...");
		picture_realloc_data(stream->picture, picture_get_generous_size(stream->dev->run->width, stream->dev->run->height));

		LOG_INFO("Capturing ...");

		while (!atomic_load(&stream->proc->stop)) {
			struct _worker_t *ready_worker;

			SEP_DEBUG('-');
			LOG_DEBUG("Waiting for worker ...");

			ready_worker = _workers_pool_wait(pool);

// drops frames at higher fps from cam; does so when the worker gets job late
//ready_worker->job_timely=true;// hack to test cam switching - doesn't work, most of the time happens in select() after the device open/stream start
			if (!ready_worker->job_failed) {
				if (ready_worker->job_timely) {
					_stream_expose_picture(stream, ready_worker->buf_index, captured_fps);
					LOG_PERF("##### Encoded picture exposed; worker=%u", ready_worker->number);
				} else {
					LOG_PERF("----- Encoded picture dropped; worker=%u", ready_worker->number);
					//LOG_INFO("----- Encoded picture dropped; worker=%u", ready_worker->number);
				}
			} else {
				break;
			}

			if (atomic_load(&stream->proc->stop)) {
				break;
			}

			if (atomic_load(&stream->proc->slowdown)) {
				usleep(1000000);
			}

                        // HTTP-initiated disconnect-reconnect to /dev/videoX
                        if(stream->reconnect){
                          break;
                        }

			bool has_read;
			bool has_write;
			bool has_error;
			// wait for a frame to read
			int selected = device_select(stream->dev, &has_read, &has_write, &has_error);

			if (selected < 0) {
				if (errno != EINTR) {
					LOG_PERROR("Mainloop select() error");
					break;
				}

			} else if (selected == 0) {
#				ifdef WITH_GPIO
				GPIO_SET_LOW(stream_online);
#				endif

				if (stream->dev->persistent) {
					if (!persistent_timeout_reported) {
						LOG_ERROR("Mainloop select() timeout, polling ...")
						persistent_timeout_reported = true;
					}

					continue;
				} else {
					LOG_ERROR("Mainloop select() timeout");
					break;
				}

			} else {
				persistent_timeout_reported = false;

				if (has_read) {
					LOG_DEBUG("Frame is ready");

#					ifdef WITH_GPIO
					GPIO_SET_HIGH(stream_online);
#					endif

					int buf_index;
					long double now = get_now_monotonic();
					long long now_second = floor_ms(now);

					if ((buf_index = device_grab_buffer(stream->dev)) < 0) {
						break;
					}

					// Workaround for broken, corrupted frames:
					// Under low light conditions corrupted frames may get captured.
					// The good thing is such frames are quite small compared to the regular pictures.
					// For example a VGA (640x480) webcam picture is normally >= 8kByte large,
					// corrupted frames are smaller.
					if (stream->dev->run->hw_buffers[buf_index].used < stream->dev->min_frame_size) {
						LOG_DEBUG("Dropped too small frame sized %zu bytes, assuming it was broken",
							stream->dev->run->hw_buffers[buf_index].used);
						goto pass_frame;
					}

					{
						if (now < grab_after) {
							fluency_passed += 1;
							LOG_VERBOSE("Passed %u frames for fluency: now=%.03Lf, grab_after=%.03Lf", fluency_passed, now, grab_after);
							goto pass_frame;
						}
						fluency_passed = 0;

						// performance profiling
						if (now_second != captured_fps_second) {
							captured_fps = captured_fps_accum;
							captured_fps_accum = 0;
							captured_fps_second = now_second;
							LOG_PERF_FPS("A new second has come; captured_fps=%u", captured_fps);
                                                        real_fps=captured_fps; // to global variable for overlay
						}
						captured_fps_accum += 1;

						long double fluency_delay = _workers_pool_get_fluency_delay(pool, ready_worker);

						grab_after = now + fluency_delay;
						LOG_VERBOSE("Fluency: delay=%.03Lf, grab_after=%.03Lf", fluency_delay, grab_after);
					}

					_workers_pool_assign(pool, ready_worker, buf_index);

					goto next_handlers; // Поток сам освободит буфер

					pass_frame:

					if (device_release_buffer(stream->dev, buf_index) < 0) {
						break;
					}
				}

				next_handlers:

				if (has_write) {
					LOG_ERROR("Got unexpected writing event, seems device was disconnected");
					break;
				}

				if (has_error) {
					LOG_INFO("Got V4L2 event");
					if (device_consume_event(stream->dev) < 0) {
						break;
					}
				}
			}
		}

                if(!stream->reconnect){
		A_MUTEX_LOCK(&stream->mutex);
		stream->online = false;
		atomic_store(&stream->updated, true);
		A_MUTEX_UNLOCK(&stream->mutex);
                }

		_workers_pool_destroy(pool);
		device_switch_capturing(stream->dev, false);
		device_close(stream->dev);

#		ifdef WITH_GPIO
		GPIO_SET_LOW(stream_online);
#		endif
	}
#ifdef WITH_BLANKBUF
    blankbuf_drawend();
#endif
#ifdef WITH_FB
    fb_close();
#endif
    LOG_INFO("Main loop terminated");
}

void stream_loop_break(struct stream_t *stream) {
	atomic_store(&stream->proc->stop, true);
}

void stream_switch_slowdown(struct stream_t *stream, bool slowdown) {
	atomic_store(&stream->proc->slowdown, slowdown);
}

static struct _workers_pool_t *_stream_init_loop(struct stream_t *stream) {
	struct _workers_pool_t *pool = NULL;

	LOG_DEBUG("%s: stream->proc->stop=%d", __FUNCTION__, atomic_load(&stream->proc->stop));

	// loop until either break happens or device is opened successfully
	while (!atomic_load(&stream->proc->stop)) {
		SEP_INFO('=');

		// /dev/videoX is opened here; NULL if failed
		if ((pool = _stream_init(stream)) != NULL) return pool;

//		if ((pool = _stream_init(stream)) == NULL) {
			stream->reconnect = false; // Shad's mod, cam switching disables this on stream loop exit; may need mutex
			stream->online = false; // Shad's mod, cam switching disables this on stream loop exit; may need mutex
			LOG_INFO("Sleeping %u seconds before new stream init ...", stream->dev->error_delay);
			device_path=stream->dev->path;
			for(unsigned int t=0;t<5*stream->dev->error_delay;t++){
				if(stream->reconnect){stream->reconnect=false;break;} // got request to change stream source, error may be gone so no more waiting
				if(stream->proc->stop){stream->reconnect=false;break;} // got request to quit
#ifdef WITH_BLANKBUF
				dev_errormsg=stream->dev->errormsg;
				blankbuf_showfb();
				blankbuf_ready=0;
				encoder_compress_buffer(stream->encoder, blankbuf_dev, 0,0);
				blankbuf_ready=1;
				atomic_store(&stream->updated, true);
#endif
			        usleep(200000);
			}
//		} else {
//			break;
//		}
	}
	return pool; // should never come here but the compiler bitches
}





static struct _workers_pool_t *_stream_init(struct stream_t *stream) {

        stream->reconnect=false; // Shad's mod

	if (device_open(stream->dev) < 0) {
		goto error;
	}
	if (device_switch_capturing(stream->dev, true) < 0) {
		goto error;
	}

	encoder_prepare(stream->encoder, stream->dev);
	return _workers_pool_init(stream);

	error:
		stream->online = false;  // Shad's mod, cam switching disables this on stream loop exit so it was moved here; may need mutex?
		atomic_store(&stream->updated, true);
		device_close(stream->dev);
		return NULL;
}





static void _stream_expose_picture(struct stream_t *stream, unsigned buf_index, unsigned captured_fps) {
	A_MUTEX_LOCK(&stream->mutex);

	picture_copy(stream->dev->run->pictures[buf_index], stream->picture);

	stream->online = true;
	stream->captured_fps = captured_fps;
	atomic_store(&stream->updated, true);

	A_MUTEX_UNLOCK(&stream->mutex);
}

static struct _workers_pool_t *_workers_pool_init(struct stream_t *stream) {
	struct _workers_pool_t *pool;

	LOG_INFO("Creating pool with %u workers ...", stream->dev->run->n_workers);

	A_CALLOC(pool, 1);

	pool->n_workers = stream->dev->run->n_workers;
	A_CALLOC(pool->workers, pool->n_workers);

	A_MUTEX_INIT(&pool->free_workers_mutex);
	A_COND_INIT(&pool->free_workers_cond);

	atomic_init(&pool->workers_stop, false);

	if (stream->dev->desired_fps > 0) {
		pool->desired_frames_interval = (long double)1 / stream->dev->desired_fps;
	}

	for (unsigned number = 0; number < pool->n_workers; ++number) {
#		define WORKER(_next) pool->workers[number]._next

		A_MUTEX_INIT(&WORKER(has_job_mutex));
		atomic_init(&WORKER(has_job), false);
		A_COND_INIT(&WORKER(has_job_cond));

		WORKER(number) = number;
		WORKER(proc_stop) = &stream->proc->stop;
		WORKER(workers_stop) = &pool->workers_stop;

		WORKER(free_workers_mutex) = &pool->free_workers_mutex;
		WORKER(free_workers) = &pool->free_workers;
		WORKER(free_workers_cond) = &pool->free_workers_cond;

		WORKER(dev) = stream->dev;
		WORKER(encoder) = stream->encoder;

		A_THREAD_CREATE(&WORKER(tid), _worker_thread, (void *)&(pool->workers[number]));

		pool->free_workers += 1;

#		undef WORKER
	}
	return pool;
}


static void _workers_pool_destroy(struct _workers_pool_t *pool) {
	LOG_INFO("Destroying workers pool ...");

	atomic_store(&pool->workers_stop, true);
	for (unsigned number = 0; number < pool->n_workers; ++number) {
#		define WORKER(_next) pool->workers[number]._next

		A_MUTEX_LOCK(&WORKER(has_job_mutex));
		atomic_store(&WORKER(has_job), true); // Final job: die
		A_MUTEX_UNLOCK(&WORKER(has_job_mutex));
		A_COND_SIGNAL(&WORKER(has_job_cond));

		A_THREAD_JOIN(WORKER(tid));
		A_MUTEX_DESTROY(&WORKER(has_job_mutex));
		A_COND_DESTROY(&WORKER(has_job_cond));

#		undef WORKER
	}

	A_MUTEX_DESTROY(&pool->free_workers_mutex);
	A_COND_DESTROY(&pool->free_workers_cond);

	free(pool->workers);
	free(pool);
}

volatile int raw_worker=0;

static void *_worker_thread(void *v_worker) {
	struct _worker_t *worker = (struct _worker_t *)v_worker;
	int rawreq=0;

	A_THREAD_RENAME("ustream:wrk%u", worker->number);
	LOG_DEBUG("Hello! I am a worker #%u ^_^", worker->number);

#	ifdef WITH_GPIO
	GPIO_INIT_PIN(workers_busy_at, worker->number);
#	endif

	while (!atomic_load(worker->proc_stop) && !atomic_load(worker->workers_stop)) {
		LOG_DEBUG("Worker %u waiting for a new job ...", worker->number);

#		ifdef WITH_GPIO
		GPIO_SET_LOW_AT(workers_busy_at, worker->number);
#		endif

		A_MUTEX_LOCK(&worker->has_job_mutex);
		A_COND_WAIT_TRUE(atomic_load(&worker->has_job), &worker->has_job_cond, &worker->has_job_mutex);
		A_MUTEX_UNLOCK(&worker->has_job_mutex);

		if (!atomic_load(worker->workers_stop)) {
#			define PICTURE(_next) worker->dev->run->pictures[worker->buf_index]->_next

			LOG_DEBUG("Worker %u handling frame from buffer %u ...", worker->number, worker->buf_index);

#			ifdef WITH_GPIO
			GPIO_SET_HIGH_AT(workers_busy_at, worker->number);
#			endif


			if(raspicam_rightmost_correction)
			    if(worker->dev->format == V4L2_PIX_FMT_YUYV)
				draw_correctrightmost(worker->dev->run->hw_buffers[worker->buf_index].data,worker->dev->width,worker->dev->height);

			// Shad's mod - check if we are handling raw frame requet, and make sure we're the only worker at a time doing so
			if(!raw_worker)rawreq=atomic_load(&worker->dev->run->raw_req);
			// raw frame requested, wait for new frame so we won't server annotated one
                        if(rawreq){
                            raw_worker=1;
  		            worker->dev->run->raw_done=false;
    	                    worker->dev->run->current_buf=worker->buf_index;
			    LOG_INFO("frame overlay skipped, requested raw frame buf %i",worker->buf_index);
  		            atomic_store(&worker->dev->run->raw_sent,true);
  		            LOG_INFO("raw frame sent");
			    // wait for the end of the overlay
                    	    for(int t=0;t<300;t++){if(atomic_load(&worker->dev->run->raw_done))break;fprintf(stderr,".");usleep(5000);}
		    	    worker->dev->run->raw_req=false;
		            worker->dev->run->raw_sent=false;
		            worker->dev->run->raw_done=false;
                            raw_worker=0;
		            LOG_INFO("raw frame received by HTTP backend");
                        }else // skip compression for the non-overlayed raw frame, trade a missed frame in MJPEG stream and local display for a lack of blinking of the overlay
			// draw overlay
			{
                    	    if(worker->dev->format == V4L2_PIX_FMT_YUYV){
   	                        worker->dev->run->raw_annotated=true;
                                device_path=worker->dev->path;
                                draw_xres=worker->dev->width;
                                draw_yres=worker->dev->height;
				// draw overlay here
//				worker->dev->run->current_buf=worker->buf_index;
                                draw_overlay(worker->dev->run->hw_buffers[worker->buf_index].data);
                    	    }
			    if (encoder_compress_buffer(worker->encoder, worker->dev, worker->number, worker->buf_index) < 0) {
			    	worker->job_failed = false;
			    }
			    COPYFB(worker->dev->run->hw_buffers[worker->buf_index].data);
			}

			if (device_release_buffer(worker->dev, worker->buf_index) == 0) {
				worker->job_start_ts = PICTURE(encode_begin_ts);
				worker->last_comp_time = PICTURE(encode_end_ts) - worker->job_start_ts;
				atomic_store(&worker->has_job, false);

				LOG_VERBOSE("Compressed new JPEG: size=%zu, time=%0.3Lf, worker=%u, buffer=%u",
					PICTURE(used), worker->last_comp_time, worker->number, worker->buf_index);
			} else {
				worker->job_failed = true;
				atomic_store(&worker->has_job, false);
			}
  	                //worker->dev->run->raw_sent=true;
  	                atomic_store(&worker->dev->run->raw_annotated,false);


#			undef PICTURE
		}

		A_MUTEX_LOCK(worker->free_workers_mutex);
		*worker->free_workers += 1;
		A_MUTEX_UNLOCK(worker->free_workers_mutex);
		A_COND_SIGNAL(worker->free_workers_cond);
	}

	LOG_DEBUG("Bye-bye (worker %u)", worker->number);
#	ifdef WITH_GPIO
	GPIO_SET_LOW_AT(workers_busy_at, worker->number);
#	endif
	return NULL;
}

static struct _worker_t *_workers_pool_wait(struct _workers_pool_t *pool) {
	struct _worker_t *ready_worker = NULL;

	A_MUTEX_LOCK(&pool->free_workers_mutex);
	A_COND_WAIT_TRUE(pool->free_workers, &pool->free_workers_cond, &pool->free_workers_mutex);
	A_MUTEX_UNLOCK(&pool->free_workers_mutex);

	if (pool->oldest_worker && !atomic_load(&pool->oldest_worker->has_job)) {
		ready_worker = pool->oldest_worker;
		ready_worker->job_timely = true;
		pool->oldest_worker = pool->oldest_worker->order_next;
	} else {
		for (unsigned number = 0; number < pool->n_workers; ++number) {
			if (
				!atomic_load(&pool->workers[number].has_job) && (
					ready_worker == NULL
					|| ready_worker->job_start_ts < pool->workers[number].job_start_ts
				)
			) {
				ready_worker = &pool->workers[number];
				break;
			}
		}
		assert(ready_worker != NULL);
		ready_worker->job_timely = false; // Освободился воркер, получивший задание позже (или самый первый при самом первом захвате)
	}
	return ready_worker;
}

static void _workers_pool_assign(struct _workers_pool_t *pool, struct _worker_t *ready_worker, unsigned buf_index) {
	if (pool->oldest_worker == NULL) {
		pool->oldest_worker = ready_worker;
		pool->latest_worker = pool->oldest_worker;
	} else {
		if (ready_worker->order_next) {
			ready_worker->order_next->order_prev = ready_worker->order_prev;
		}
		if (ready_worker->order_prev) {
			ready_worker->order_prev->order_next = ready_worker->order_next;
		}
		ready_worker->order_prev = pool->latest_worker;
		pool->latest_worker->order_next = ready_worker;
		pool->latest_worker = ready_worker;
	}
	pool->latest_worker->order_next = NULL;

	A_MUTEX_LOCK(&ready_worker->has_job_mutex);
	ready_worker->buf_index = buf_index;
	atomic_store(&ready_worker->has_job, true);
	A_MUTEX_UNLOCK(&ready_worker->has_job_mutex);
	A_COND_SIGNAL(&ready_worker->has_job_cond);

	A_MUTEX_LOCK(&pool->free_workers_mutex);
	pool->free_workers -= 1;
	A_MUTEX_UNLOCK(&pool->free_workers_mutex);

	LOG_DEBUG("Assigned new frame in buffer %u to worker %u", buf_index, ready_worker->number);
}

static long double _workers_pool_get_fluency_delay(struct _workers_pool_t *pool, struct _worker_t *ready_worker) {
	long double approx_comp_time;
	long double min_delay;

	approx_comp_time = pool->approx_comp_time * 0.9 + ready_worker->last_comp_time * 0.1;

	LOG_VERBOSE("Correcting approx_comp_time: %.3Lf -> %.3Lf (last_comp_time=%.3Lf)",
		pool->approx_comp_time, approx_comp_time, ready_worker->last_comp_time);

	pool->approx_comp_time = approx_comp_time;

	min_delay = pool->approx_comp_time / pool->n_workers; // Среднее время работы размазывается на N воркеров

	if (pool->desired_frames_interval > 0 && min_delay > 0 && pool->desired_frames_interval > min_delay) {
		// Искусственное время задержки на основе желаемого FPS, если включен --desired-fps
		return pool->desired_frames_interval;
	}
	return min_delay;
}
